//---------------------------------------------------------------------------
//
//  FILE: dinero.C:
//  Author: Ken Barr
//  Date: Mon Oct 31 11:32:48 2005
//  Description:  A Pintool that feeds the Dinero cache simulator
//  Copyright (C) 2005 Kenneth Barr (see COPYRIGHT file for details)
//
//---------------------------------------------------------------------------


#include <stdio.h>  //pin chokes if you use <cstdio>
#include <iostream>
#include "pin.H"
#include "instlib.H"
#include "d4.h"
#include "cmdd4.h"
#include "cmdargs.h"
#include "tracein.h"
#include "inttypes.h"

void doargs (int, char **);
void dostats (FILE * );
void summarize_caches (FILE *, d4cache *, d4cache *);
void finalize();


FILE * outfile;
d4memref r;
d4cache *ci, *cd;
double tmaxcount = 0, tintcount;
double flcount;
unsigned long long cpu_bytes_read=0, cpu_bytes_written=0;
UINT64 num_insn=0;


KNOB<string> KnobOutputFile(KNOB_MODE_WRITEONCE,    "pintool",
    "f", "dinero.out", "specify dinero output file name");
KNOB<string> KnobDineroCfgFile(KNOB_MODE_WRITEONCE,    "pintool",
    "cfg", "/home/bits8/kbarr/acip/config/dcache_8B_lines.cfg", "specify dinero config input file name");
KNOB<UINT32> KnobFinalFlush(KNOB_MODE_WRITEONCE,    "pintool",
    "flush", "0", "writeback dirty lines upon completion");
KNOB<UINT32> KnobTraceData(KNOB_MODE_WRITEONCE,    "pintool",
    "tracedata", "1", "pass data references to dinero");
KNOB<UINT32> KnobTraceInst(KNOB_MODE_WRITEONCE,    "pintool",
    "traceinst", "0", "pass inst references to dinero");


/* ------------------------------------------------------------------------
   Controller Stuff -- Thanks to Harish for explaining this to me
   -----------------------------------------------------------------------*/
bool do_work=false;
INSTLIB::CONTROL control;

VOID Handler(INSTLIB::CONTROL_EVENT ev, VOID * v, VOID * ctxhandle, VOID * ip, VOID * tid)
{
    switch(ev)
    {
      case INSTLIB::CONTROL_START:
        cout << "Starting instrumentation..." << endl;
        do_work = true;
        PIN_RemoveInstrumentation();  //actually *refreshes* the instrumentation functions
        break;

      case INSTLIB::CONTROL_STOP:
// if we reach this point and the program is already done, Pin won't do the RemoveFini/Detach.
// This is what happens when I'm not using the control switches.
        cout << "Stopping Pin..." << endl;
        PIN_RemoveFiniFunctions();
        PIN_Detach();
        finalize();
        break;

      default:
        ASSERTX(false);
        break;
    }
}


VOID CountInst()
{
    num_insn++;
}


//detach when done (useful to make sure program ran correctly)
VOID docount_internal()
{
        PIN_RemoveFiniFunctions();
        PIN_Detach();
}

// hard exit when done
VOID docount_internal1()
{
    finalize();
    exit(0);  
}

VOID docount()
{
    if(KnobTraceInst || KnobTraceData)
    {
        tmaxcount += 1;
        if ((maxcount  > 0) && (tmaxcount >= maxcount))
        {
            fprintf (outfile, "---Maximum address count exceeded.\n");
            docount_internal1();
        }
    }
}

VOID periodic()
{
    if (tintcount > 0 && (tintcount -= 1) <= 0)
    {
        dostats(outfile);
        tintcount = stat_interval;
    }

    if (flcount > 0 && (flcount -= 1) <= 0)
    {
        // flush cache = copy back and invalidate
        r.accesstype = D4XCOPYB;
        r.address = 0;
        r.size = 0;
        d4ref (cd, r);
        r.accesstype = D4XINVAL;
        d4ref (ci, r);
        if (ci != cd)
        {
            d4ref (cd, r);
        }
        flcount = flushcount;
    }
}

// wrap up the data and call dinero
VOID CallDinero(VOID * addr, VOID * size, char accesstype)
{
    static char buffer[512];
    asm("fxsave %0" : "=m"(buffer));  // work around a FP problem in Pin

    // Pin/x86 reports some memory instructions (REP MOV) as 0-length.  
    // Dinero can't handle this. 
    // XXX: need to handle variable memory reference size and instructions with >1 memory operand
    if((uint32_t)size > 0)            
    {

        r.address = (uint32_t)addr;
        r.size = (uint32_t)size;
        r.accesstype = accesstype;

//      fprintf(stderr, "%.0f read   %8x. size %d\n", tmaxcount, addr, size);

        docount();      // check for an early exit
        if(accesstype == D4XINSTRN)
        {
            d4ref (ci, r);  // call dinero
        }
        else
        {
            if(accesstype == D4XREAD)
            {
                cpu_bytes_read += (uint32_t)size;
            }
            else
            {
                cpu_bytes_written += (uint32_t)size;
            }
            d4ref (cd, r);
        }
        periodic();     // dinero supports occasional flushing, stats
    }
    asm("fxrstor %0" : "=m"(buffer));

}


// Is called for every instruction and instruments reads and writes
VOID Instruction(INS ins, VOID *v)
{

    /* the calls below won't be inserted unless do_work is true.
       This helps us get to the interesting point faster. */
    if(do_work)
    {
        INS_InsertPredicatedCall(
            ins, IPOINT_BEFORE, (AFUNPTR)CountInst,
            IARG_INST_PTR,
            IARG_END);

        // every instruction accesses the i-cache
        if(KnobTraceInst)
        {
            INS_InsertPredicatedCall(
                ins, IPOINT_BEFORE, (AFUNPTR)CallDinero,
                IARG_INST_PTR,
                IARG_UINT32, INS_Size (ins),
                IARG_UINT32, D4XINSTRN,
                IARG_END);
        }

        if(KnobTraceData)
        {
            if (INS_IsMemoryRead(ins))
            {
                INS_InsertPredicatedCall(
                    ins, IPOINT_BEFORE, (AFUNPTR)CallDinero,
                    IARG_MEMORYREAD_EA,
                    IARG_MEMORYREAD_SIZE,
                    IARG_UINT32, D4XREAD,
                    IARG_END);

                //XXX: handle the 2-argument instructions
            }


            // instruments stores using a predicated call, i.e.
            // the call happens iff the store will be actually executed
            if (INS_IsMemoryWrite(ins))
            {
                INS_InsertPredicatedCall(
                    ins, IPOINT_BEFORE, (AFUNPTR)CallDinero,
                    IARG_MEMORYWRITE_EA,
                    IARG_MEMORYWRITE_SIZE,
                    IARG_UINT32, D4XWRITE,
                    IARG_END);
            }
        }
    }
}


void finalize()
{
    static bool done= false;  //don't finalize twice!
    if(!done)
    {
        fprintf(stderr, "finalize!!\n");
        if(KnobFinalFlush == true)
        {
            r.accesstype = D4XCOPYB;
            r.address = 0;
            r.size = 0;
            d4ref (cd, r);
        }

        fprintf(outfile,  "---Simulation complete.\n");

        fprintf(outfile, "Bytes read by CPU: %llu\n", cpu_bytes_read);
        fprintf(outfile, "Bytes written by CPU: %llu\n", cpu_bytes_written);
        fprintf(outfile, "Insts executed: %llu\n", num_insn);
        dostats(outfile);
        fprintf(outfile,  "---Execution complete.\n");
        fclose(outfile);
        done = true;
    }

}


VOID Fini(INT32 code, VOID *v)
{
    fprintf(stderr, "fini!!\n");
    fprintf(outfile, "fini!!\n");
    finalize();
}

VOID Detach(VOID *v)
{
    fprintf(stderr, "detach!!\n");
    fprintf(outfile, "detach!!\n");
    finalize();
}





int main(int argc, char *argv[])
{

    if( PIN_Init(argc,argv) )
    {
        cerr << KNOB_BASE::StringKnobSummary();
        return -1;
    }



  FILE * cfgfile = fopen(KnobDineroCfgFile.Value().c_str(), "r");
#define LINESIZE 80
#define MAXARGS 50
  char line[LINESIZE];
  int linelength;
  int lib_argc=0;
  char * lib_argv[MAXARGS];

  if(cfgfile == NULL)
  {
      cout << "Could not open " << KnobDineroCfgFile.Value().c_str() << endl;
  }
  lib_argv[lib_argc++] = strdup("dinero");
  while(fgets(line, LINESIZE, cfgfile))
  {
      linelength = strlen(line);
      if(line[linelength-1] == '\n')
      {
          line[linelength-1] = '\0';  //replace newline with NULL
      }
      lib_argv[lib_argc++] = strdup(line);
  }
  cout << "dinero opts: ";
  for(int i=0; i<lib_argc; i++)
  {
      cout << lib_argv[i] << " ";
  }
  cout << endl;

  if(strcmp(KnobOutputFile.Value().c_str(), "none") == 0)
  {
      outfile = stdout;
  }
  else{
      outfile = fopen(KnobOutputFile.Value().c_str(), "w");
  }

  doargs (lib_argc, lib_argv);
  verify_options();
  initialize_caches (&ci, &cd);
  summarize_caches (outfile, ci, cd);

  fprintf (outfile, "\n---Simulation begins.\n");
  fprintf (outfile, "\nmaxcount is: %f\n", maxcount);

  tintcount = stat_interval;
  flcount = flushcount;

  // Activate alarm, must be done before PIN_StartProgram
  control.CheckKnobs(Handler, 0);

  INS_AddInstrumentFunction(Instruction, 0);
  PIN_AddDetachFunction(Detach, 0);
  PIN_AddFiniFunction(Fini, 0);


  // Never returns
  PIN_StartProgram();

  return 0;
}



